
#ifndef __UPF_CXX_API_H__
#define __UPF_CXX_API_H__

namespace upf
{

// -----------------------------------------------------------------------------------
// upf client-side API:
// -----------------------------------------------------------------------------------


#ifndef UPF_BUILDING_MODULE
/** Initializes UPF. Must be called exactly once by the application.
    \return true on success, false on failure
    \note Must \b not be called from UPF module DLL! Call it only from the app.
    \see upf::done
 */
inline bool init()
    { return (bool)upf_init(); }

/** Shuts down UPF. Must be called exactly once by the application unless
    upf::init failed.
    \return true on success, false on failure
    \note Must \b not be called from UPF module DLL! Call it only from the app.
    \see upf::init
 */
inline bool done()
    { return (bool)upf_done(); }
#endif

/** Returns pointer to the IManager singleton. IManager instance is used to access
    UPF's global state (e.g. class info and type info repositories, IID to interface name
    mappings etc.).
    \see upf::IManager
 */
inline Ptr<IManager> getManager()
    { return _makeCxxObject<IManager>(upf_middleware->manager); }

/** Returns pointer to the IServiceManager singleton. IServiceManager instance
    is used to access UPF's services (i.e. named objects).
    \see upf::IServiceManager
 */
inline Ptr<IServiceManager> getServiceManager()
    { return _makeCxxObject<IServiceManager>(upf_middleware->serviceManager); }

/** Translates string class name to CID identifier.
    \return CID value or 0 if no such class is registered.
 */
inline CID getCID(const char *classname)
    { return upf_middleware->funcs.classname_to_cid(classname); }

/** Creates instance of class with CID equal to \a classID. You can obtain
    \a classID by calling getCID, or you can pass the classname as argument
    to create.
    \note This is slightly more efficient that passing \a classname as argument
          to the other prototype off upf::create
 */
inline Ptr<IObject> create(CID classID)
{
    Ptr<IObject> ptr = _makeCxxObject<IObject>(upf_middleware->funcs.create_instance(classID));
    if (ptr)
    {
        // compensate for incRef in funcs.create_instance:
        ptr->_upf_getABIObject()->refCnt--; 
    }
    return ptr;
}

/// Creates instance of class with given class name.
inline Ptr<IObject> create(const char *classname)
    { return create(getCID(classname)); }

/// Creates instance of class with given class name.
inline Ptr<IObject> create(const std::string& classname)
    { return create(getCID(classname.c_str())); }

/** Creates instance of object implementing \a Interface. This prototype
    is only available for interfaces that were defined with #pragma single_impl
    in IDL file.

    \see \ref pragma_single_impl
 */

#ifndef _MSC_VER
template<class Interface> inline Ptr<IObject> create()
#else
// NB: dummy parameter workarounds a bug in VisualC++
template<class Interface> inline Ptr<IObject> create(Interface*** = NULL) 
#endif
{
    return create(InterfaceTraits<Interface>::implName());
}

}


#endif
