
#include "upf/upf.h"
#include "private.h"
#include <algorithm>

namespace upf { namespace impl {

// --------------------------------------------------------------------------
// IWriteableClassInfo part (write-only, structure initialization)
// --------------------------------------------------------------------------

void ClassInfo::_initInstance()
{
    _upf_initInstance();
}

void ClassInfo::addInterface(const IID& id)
{
    assert(!m_frozen);
    if (find(m_ifaces.begin(), m_ifaces.end(), id) == m_ifaces.end())
    {
        m_ifaces.push_back(id);
    }
}

void ClassInfo::setCID(CID id)
{
    assert(!m_frozen);
    m_cid = id;
}

void ClassInfo::setName(const string& name)
{
    assert(!m_frozen);
    m_name = name;
}

void ClassInfo::setFactory(IClassFactory *f)
{
    assert(!m_frozen);
    m_factory = f;
}

void ClassInfo::addProperty(const string& name, const string& value)
{
    assert(!m_frozen);
    m_properties[name] = value;
}

Ptr<IClassInfo> ClassInfo::freeze()
{
    m_frozen = true;
    return this;
}


// --------------------------------------------------------------------------
// IClassInfo part (read-only)
// --------------------------------------------------------------------------

bool ClassInfo::implementsInterface(const IID& iface)
{
    return find(m_ifaces.begin(), m_ifaces.end(), iface) != m_ifaces.end();
}

string ClassInfo::getProperty(const string& name)
{
    StringHash<string>::iterator i = m_properties.find(name);
    if (i == m_properties.end())
        return "";
    else
        return i->second;
}

ClassInfo::Properties ClassInfo::getProperties()
{
    Properties list;
    Property prop;
    for (StringHash<string>::iterator i = m_properties.begin(); i != m_properties.end(); i++)
    {
        prop.name = i->first;
        prop.value = i->second;
        list.push_back(prop);
    }
    return list;
}




} }

UPF_IMPLEMENT_CLASS(upf::impl::ClassInfo)
{
    UPF_INTERFACE(upf::IClassInfo)
    UPF_INTERFACE(upf::IWriteableClassInfo)
    UPF_PROPERTY("PartOf",      "UPF core")
    UPF_PROPERTY("Author",      "Vaclav Slavik")
}
