
#ifndef __UPF_CLASSINFO_H__
#define __UPF_CLASSINFO_H__

#include "upf/upf.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

namespace upf { namespace impl {

using namespace std;

class Manager;


class ClassInfo : public IClassInfo, public IWriteableClassInfo
{
    friend class Manager;
    
public:
    ClassInfo() : m_frozen(false) {}
    virtual ~ClassInfo() {}

    void _initInstance();

    // Read-only part:
    CID getCID() { return m_cid; }
    string getName() { return m_name; }
    Ptr<IClassFactory> getFactory() { return m_factory; }
    IIDsList getInterfaces() { return m_ifaces; }
    bool implementsInterface(const IID& iface);
    string getProperty(const string& name);
    Properties getProperties();

    // Writeable part:
    void setCID(CID id);
    void setName(const string& name);
    void setFactory(IClassFactory *f);
    void addInterface(const IID& id);
    void addProperty(const string& name, const string& value);
    
    Ptr<IClassInfo> freeze();
    
private:
    bool               m_frozen;
    string             m_name;
    Ptr<IClassFactory> m_factory;
    IIDsList           m_ifaces;
    CID                m_cid;
    StringHash<string> m_properties;

    UPF_DECLARE_CLASS(ClassInfo)
};


} } // end of namespace

#endif
