
#include <upf/upf.h>
#include <upf/ILoader.h>
#include "direnum.h"

using namespace std;

#include <stdlib.h>

#ifdef __UPF_UNIX__
#include <dirent.h>

namespace upf { namespace impl {

void enumDirectory(const std::string& path,
                   const std::vector<std::string>& extensions,
                   std::vector<std::string>& files)
{
    DIR   *dir;
    std::vector<std::string>::const_iterator ext;
    bool allFiles = (extensions.size() == 0);
    
    files.clear();
    
    dir = opendir(path.c_str());
    if (dir)
    {
        struct dirent *dentry;

        for (dentry = readdir(dir); dentry; dentry = readdir(dir))
        {
            size_t nlen = strlen(dentry->d_name);
            if (nlen <= 2 && 
                (strcmp(dentry->d_name, ".") == 0 || strcmp(dentry->d_name, "..") == 0))
                continue;
            
            if (allFiles)
            {
                files.push_back(path + "/" + dentry->d_name);
            }
            else
            {
                for (ext = extensions.begin(); ext != extensions.end(); ext++)
                {
                    string stable(*ext);
                    const char *ex = stable.c_str();
                    if (strcmp(dentry->d_name + nlen - ext->length(), ex) == 0)
                        files.push_back(path + "/" + dentry->d_name);
                }
            }
        }
        closedir(dir);
    }
}

} }
#endif

#ifdef __UPF_WIN32__
#include <windows.h>

namespace {

void doEnum(const std::string& path,
            const std::string& wildcard, std::vector<std::string>& files)
{
    WIN32_FIND_DATA data;
    HANDLE hFind;

    hFind = FindFirstFile(wildcard.c_str(), &data);
    while (hFind != INVALID_HANDLE_VALUE)
    {
        if ((data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
        {
            files.push_back(path + "\\" + data.cFileName);
        }
        if (!FindNextFile(hFind, &data)) break;
    }
    FindClose(hFind);
}

}

namespace upf { namespace impl {

void enumDirectory(const std::string& path,
                   const std::vector<std::string>& extensions,
                   std::vector<std::string>& files)
{
    string wildcard;
    files.clear();

    if (extensions.size() == 0)
    {
        wildcard = path + "\\*.*";
        doEnum(path, wildcard, files);
    }
    else
    {
        for (std::vector<std::string>::const_iterator ext = extensions.begin();
             ext != extensions.end(); ext++)
        {
            wildcard = path + "\\*" + *ext;
            doEnum(path, wildcard, files);
        }
    }
}

} }
#endif
