
#include <upf/upf.h>
#include <upf/ILoader.h>

#include <stdlib.h>

#include "private.h"
#include "dll_loader.h"

#ifdef __UPF_UNIX__
#include "upf_dlfcn.h"
#endif

using namespace std;
using namespace upf;
using namespace upf::impl;

namespace {

_upf_dllEntryPointFunc resolveEntryPoint(DllModule& dll)
{
    _upf_dllEntryPointFunc func; 
    func = (_upf_dllEntryPointFunc)
           dll.getSymbol(_UPF_DLL_ENTRY_POINT_NAME);
    if (func == NULL)
    {
        func = (_upf_dllEntryPointFunc)
               dll.getSymbol("_" _UPF_DLL_ENTRY_POINT_NAME);
    }
    return func;
}
	
}

namespace upf { namespace impl {


#ifdef __UPF_UNIX__

bool DllModule::load(const string& filename)
{
    m_handle = dlopen(filename.c_str(), RTLD_NOW | RTLD_LOCAL | RTLD_GROUP);
    return m_handle != NULL;
}

void *DllModule::getSymbol(const string& name)
{
    return dlsym(m_handle, name.c_str());
}

void DllModule::unload()
{
    dlclose(m_handle);
}
#endif

#ifdef __UPF_WIN32__
bool DllModule::load(const string& filename)
{
    m_handle = LoadLibrary(filename.c_str());
    return m_handle != 0;
}

void *DllModule::getSymbol(const string& name)
{
    return (void*)GetProcAddress(m_handle, name.c_str());
}

void DllModule::unload()
{
    FreeLibrary(m_handle);
}
#endif


DllLoader::DllLoader()
{
}

DllLoader::~DllLoader()
{
    _upf_dllEntryPointFunc func;
    for (ModulesList::iterator i = m_modules.begin(); 
         i != m_modules.end(); i++)
    {
        func = resolveEntryPoint(*i);
        if (func != NULL)
            func(upf_middleware, _UPF_DLL_ACTION_UNLOAD);
        i->unload();
    }
}

ILoader::ExtensionsList DllLoader::getExtensions()
{
    ExtensionsList list;
#ifdef __UPF_UNIX__
    list.push_back(".so");
#endif
#ifdef __UPF_WIN32__
    list.push_back(".dll");
#endif
    return list;
}

bool DllLoader::canLoad(const string&)
{
    return true;
}

bool DllLoader::load(const string& filename)
{
    _upf_dllEntryPointFunc func;
    DllModule dll;

    if (dll.load(filename))
    {
        func = resolveEntryPoint(dll);
        if (func != NULL)
        {
            if (!func(upf_middleware, _UPF_DLL_ACTION_LOAD) ||
                !func(upf_middleware, _UPF_DLL_ACTION_INIT_CLASSES))
            {
                privateData->log->log(
                          "upf", ILog::LOG_ERROR,
                          string("Failed to load DLL module: ") + filename);
            }
            else
            {
                m_modules.push_back(dll);
                return true;
            }
        }
        // failure
        dll.unload();
    }
    return false;
}

string DllLoader::getDescription()
{
   return "shared library";
}


} } // namespace upf

UPF_IMPLEMENT_CLASS(upf::impl::DllLoader)
{
    UPF_INTERFACE(upf::ILoader)
    UPF_PROPERTY("PartOf",      "UPF core")
    UPF_PROPERTY("Author",      "Vaclav Slavik")
#ifdef __UPF_UNIX__
    UPF_PROPERTY("Description", "Shared libraries loader")
#else
    UPF_PROPERTY("Description", "DLLs loader")
#endif
}
