
#ifndef __UPF_HASHES_H__
#define __UPF_HASHES_H__

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined(HAVE_EXT_HASH_MAP)
  #include <ext/hash_map>
  #ifdef __GNUC__
    using namespace __gnu_cxx;
  #endif
  #define USE_HASH
#elif defined(HAVE_HASH_MAP)
  #include <hash_map>
  #define USE_HASH
#else
  #include <map>
#endif


#ifdef USE_HASH
  #define best_map_type hash_map
#else
  #define best_map_type map
#endif


namespace upf { namespace impl {

using namespace std;

// ---------------------------------------------------------------------------
// string maps
// ---------------------------------------------------------------------------

#ifdef USE_HASH
struct eqstr
{   
    bool operator()(const string& s1, const string& s2) const
    {
        return s1 == s2;
    }
};
    
struct stringhash
{
    hash<const char*> impl;
    size_t operator()(const string& s) const { return impl(s.c_str()); }
};

template<class T> class StringHash : public hash_map<string, T, stringhash, eqstr> {};

#else
struct ltstr
{   
    bool operator()(const string& s1, const string& s2) const
    {
        return s1.compare(s2) < 0;
    }
};

template<class T> class StringHash : public std::map<std::string, T, ltstr> {};
#endif






// ---------------------------------------------------------------------------
// upf::IID maps
// ---------------------------------------------------------------------------

#ifdef USE_HASH

struct IID_hash
{
  size_t operator()(const upf::IID& x) const
  {
      return x.data1+x.data2+x.data3+x.data4+x.version_major+x.version_minor;
  }
};

template<class T> class IIDHash : public hash_map< ::upf::IID, T, IID_hash> {};

#else

struct IID_lt
{   
    bool operator()(const upf::IID& i1, const upf::IID& i2) const
    {
        #define __UPF_CMP(field) \
             if (i1.field != i2.field) return i1.field < i2.field
        __UPF_CMP(data1);
        __UPF_CMP(data2);
        __UPF_CMP(data3);
        __UPF_CMP(data4);
        __UPF_CMP(version_major);
        __UPF_CMP(version_minor);
        #undef __UPF_CMP
        return false;
    }
};

template<class T> class IIDHash : public map< ::upf::IID, T, IID_lt> {};

#endif



} } // end of namespace

#endif
