
#ifndef __UPF_MANAGER_H__
#define __UPF_MANAGER_H__

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "upf/upf.h"
#include "upf/ILoader.h"

#include <list>

namespace upf { namespace impl {

using namespace std;

class Manager : public IManager
{
public:
    Manager();
    virtual ~Manager();
    
    void _initClass();

    bool registerClass(IClassInfo *classinfo);
    Ptr<IObject> createInstance(CID classID);
    CIDsList enumInterfaceImplementations(const IID& iface);
    void addModulesPath(const string& path);
    bool loadModule(const string& filename);

    Ptr<IClassInfo> getClassInfo(CID classID);
    Ptr<IInterfaceInfo> getInterfaceInfo(const IID& iid);

    CID classname2cid(const char *classname);

    void addIIDMappings(const IIDMappingData& m);
    
private:
    void addLoader(ILoader *loader);
    void loadClasses(ILoader *loader, const string& path);

    struct ClassInfoData
    {
        Ptr<IClassInfo>    classInfo;
        Ptr<IClassFactory> factory;
    };

    typedef best_map_type<CID, ClassInfoData> ClassesTbl;
    typedef StringHash<CID> ClassnamesTbl;
    typedef IIDHash<string> IIDsTbl;
    typedef list< Ptr<ILoader> > LoadersList;

    ClassesTbl m_classes;
    ClassnamesTbl m_classnames;
    IIDsTbl m_iids;
    
    list<string> m_modulesPaths;
    LoadersList m_loaders;

    UPF_DECLARE_CLASS(Manager)
};


} } // end of namespace

#endif
