
#ifndef __UPF_PRIVATE_H__
#define __UPF_PRIVATE_H__

#include "upf/upf.h"
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined(__UPF_UNIX__) && !defined(HAVE_PTHREAD)
    #define BOOST_NO_MT
#endif

#include "boost/pool/pool.hpp"
#include "hashes.h"
#include "classinfo.h"
#include "typeinf.h"
#include "manager.h"
#include "upf/ILog.h"
#include "services.h"
#include "logging.h"


namespace upf { namespace impl {

using namespace std;

upf_Object *create_instance(upf_CID classID);
void marchaller_init(upf_Arguments *args);
void marchaller_seek0(upf_Arguments *args);
void marchaller_destroy(upf_Arguments *args);
void marchaller_write(upf_Arguments *args, void *data, size_t len);
void marchaller_read(upf_Arguments *args, void *data, size_t len);
upf_ContextID get_context(const char *context_string);
void init_std_contexts(const char *user_context);
upf_CID alloc_new_cid(void);
upf_CID classname_to_cid(const char *classname);
int merge_typeinfo(const upf_uint8_t *typeinf);

struct PrivateData
{
    PrivateData() : 
            manager(NULL), argsPool(sizeof(upf_ArgumentsChunk)) {}

    Manager                  *manager;
    Ptr<IServiceManager>      serviceManager;
    boost::pool<>             argsPool;
    StringHash<upf_ContextID> contexts;
    TypeInfoRegistry          typeInfoReg;
    Ptr<ILog>                 log;
};

extern PrivateData *privateData;


} } // end of namespace

#endif
