
/*

   This little beats is only needed on Unix platforms. The problem that
   we are facing here is that Python (shared) modules are compiled with
   unresolved symbols on Unix and are resolved at runtime from global
   symbols namespace. This is usually OK because Python interpreter or
   apps that embed Python link against libpython. In UPF, however, only
   python-loader.so needs libpython and because python-loader.so is loaded
   using dlopen() with RTLD_LOCAL flag, its symbols and symbols from all
   libraries it links against are *not* imported into global namespace.
   Consequently, Python is unable to load its modules because of unresolved
   symbols.
   
   In order to solve this, we use python-loader-helper.so, which is a small
   shared library that is loaded at runtime by python-loader.so with
   RTLD_GLOBAL flag. This way all symbols from python-loader-helper.so are
   imported into global namespace and because the helper links against
   libpython, all Python symbols are available, and C++ symbols from
   python-loader.so are not globally visible.

 */

#include "config.h"
#include PYTHON_INCLUDE

#ifdef __cplusplus
extern "C" {
#endif

typedef struct
{
    DL_IMPORT(void) (*Py_Initialize)(void);
    DL_IMPORT(void) (*Py_Finalize)(void);
    DL_IMPORT(int)  (*Py_IsInitialized)(void);
    DL_IMPORT(int)  (*PyRun_SimpleString)(char *);
    DL_IMPORT(int)  (*PyRun_SimpleFile)(FILE *, char *);
} upf_python_api;

extern upf_python_api *__upf_getPythonAPI(void);
typedef upf_python_api* (*__upf_getPythonAPI_Func)(void);
#define __upf_getPythonAPI_Name "__upf_getPythonAPI"

#ifdef __cplusplus
}
#endif
