
import struct

def encUint8(data):
    return struct.pack('!B', data)

def encBool(data):
    if data:
        return encUint8(1)
    else:
        return encUint8(0)

def encUint16(data):
    return struct.pack('!H', data)

def encUint32(data):
    return struct.pack('!L', data)

def encString(data):
    x = str(data)
    return struct.pack('!H%is' % len(x), len(x), x)

def encIID(IID, version):
    major=int(version.split('.')[0])
    minor=int(version.split('.')[1])
    iid1=long("0x" + IID[0:8], 0)
    iid2=long("0x" + IID[8:16], 0)
    iid3=long("0x" + IID[16:24], 0)
    iid4=long("0x" + IID[24:32], 0)
    return struct.pack('!LLLLHH', iid1, iid2, iid3, iid4, major, minor)




# Entity types:
ENTITY_INTERFACE   = 0
ENTITY_TYPEDEF     = 1
ENTITY_NATIVE_TYPE = 2
ENTITY_STRUCT      = 3
ENTITY_CONST       = 4
ENTITY_UNION       = 5
ENTITY_ENUM        = 6

TYPEDEF_SEQUENCE   = 1
TYPEDEF_SIMPLE     = 2

class MethodParameter:
    def __init__(self, type, name, isIn, isOut):
        self.type = type
        self.name = name
        self.isIn = isIn != 0 and isIn != '0'
        self.isOut = isOut != 0 and isOut != '0'

class Method:
    def __init__(self, id, returns, params):
        self.id = id
        self.returns = returns
        self.params = params

class Variable:
    def __init__(self, type, name):
        self.type = type
        self.name = name

class UnionCase:
    def __init__(self, value, type, name):
        self.value = value
        self.type = type
        self.name = name
        


def serializeCommon(obj):
    assert obj.ent_type != None
    return encUint8(obj.ent_type) + encString(str(obj.id))


def serializeNativeType(obj):
    return serializeCommon(obj)


def serializeTypedef(obj):
    return serializeCommon(obj) + \
           encUint8(obj.kind) + encString(obj.type)


def serializeConst(obj):
    return serializeCommon(obj) + \
           encString(obj.type) + encString(obj.value)


def serializeInterface(obj, allEnts):
    s = serializeCommon(obj) + encIID(obj.IID, obj.version)

    s += encUint16(len(obj.super))
    for sup in obj.super:
        s += encIID(allEnts[sup].IID, allEnts[sup].version)

    def serializeMethod(m):
        x = encString(m.id.split('.')[-1])
        x += encString(m.returns)
        x += encUint16(len(m.params))
        for p in m.params:
            x += encString(p.type)
            x += encString(p.name)
            x += encBool(p.isIn)
            x += encBool(p.isOut)
        return x

    s += encUint16(len(obj.methods))
    for m in obj.methods:
        s += serializeMethod(m)

    s += encBool(obj.singleImpl)
    return s


def serializeStruct(obj):
    s = serializeCommon(obj)
    s += encUint16(len(obj.variables))
    for v in obj.variables:
        s += encString(v.type) + encString(v.name)
    return s


def serializeEnum(obj):
    s = serializeCommon(obj)
    for v in obj.values:
        s += encString(v.split('.')[-1])
    return s


def serializeUnion(obj):
    s = serializeCommon(obj) + encString(obj.switchType)
    s += encUint16(len(obj.cases))
    for c in obj.cases:
        s += encString(c.value) + encString(c.type) + encString(c.name)
    return s
