
#include "upf/upf.h"
#include "private.h"

namespace upf { namespace impl {


ServiceManager::ServiceManager()
{
}

ServiceManager::~ServiceManager()
{
}

bool ServiceManager::bind(const string& name, IObject* obj, long permissions)
{
    assert(obj != NULL);
    assert(!name.empty());

    ServicesTbl::iterator entry = m_services.find(name);
    if (entry == m_services.end())
    {
        entry = m_services.insert(
                    ServicesTbl::value_type(name, ServiceData())).first;
        return doBind(entry, obj, permissions);
    }
    else
    {
        string err = "Service already registered: " + name;
        privateData->log->log("upf", ILog::LOG_ERROR, err);
        return false;
    }
}

bool ServiceManager::rebind(const string& name, IObject* obj)
{
    assert(obj != NULL);

    ServicesTbl::iterator entry = m_services.find(name);
    if (entry == m_services.end())
    {
        string err = "Service not found: " + name;
        privateData->log->log("upf", ILog::LOG_ERROR, err);
        return false;
    }
    else if ((entry->second.permissions & SERVICE_CAN_REBIND) == 0)
    {
        string err = "No permissions to rebind service: " + name;
        privateData->log->log("upf", ILog::LOG_ERROR, err);
        return false;
    }
    else
    {
        return doBind(entry, obj, entry->second.permissions);
    }
}

bool ServiceManager::unbind(const string& name)
{
    ServicesTbl::iterator entry = m_services.find(name);
    if (entry == m_services.end())
    {
        string err = "Service not found: " + name;
        privateData->log->log("upf", ILog::LOG_ERROR, err);
        return false;
    }
    else if ((entry->second.permissions & SERVICE_CAN_UNBIND) == 0)
    {
        string err = "No permissions to unbind service: " + name;
        privateData->log->log("upf", ILog::LOG_ERROR, err);
        return false;
    }
    else
    {
        doBind(entry, NULL, entry->second.permissions);
        if (entry->second.object == NULL) // no listener re-added it
            m_services.erase(entry);
        return true;
    }
}

bool ServiceManager::doBind(ServicesTbl::iterator& service,
                            IObject *object, long perm)
{
    service->second.object = object;
    service->second.permissions = perm;

    ListenersTbl::iterator ls = m_listeners.find(service->first);
    if (ls != m_listeners.end())
    {
        Listeners& list = ls->second;
        for (Listeners::iterator l = list.begin(); l != list.end(); l++)
            (*l)->onChange(service->first, object);
    }
    return true;
}

Ptr<IObject> ServiceManager::resolve(const string& name)
{
    ServicesTbl::iterator entry = m_services.find(name);
    if (entry == m_services.end())
        return (IObject*)NULL;
    else
        return entry->second.object;
}

void ServiceManager::addListener(const string& name, IServiceChangeListener* listener)
{
    m_listeners[name].push_back(listener);
}
    



UPF_IMPLEMENT_CLASS(upf::impl::ServiceManager)
{
    UPF_INTERFACE(upf::IServiceManager)
    UPF_PROPERTY("PartOf",      "UPF core")
    UPF_PROPERTY("Author",      "Vaclav Slavik")
}


} }
