
#include <stdio.h>

#include "config.h"
#include "upf/upf.h"
#include "upf_dlfcn.h"

#define UPF_CORE_NAME "libupf_impl-" PACKAGE_VERSION ".so"

static void *handle = 0;
static int (*init)(const char *app_context,
                   upf_ContextID *context, upf_middlewareLayer **middleware);
static int (*done)(void);

upf_middlewareLayer *upf_middleware = NULL;
upf_ContextID upf_context = 0;

static void report_dlerror()
{
    fprintf(stderr, "can't load %s: %s\n", UPF_CORE_NAME, dlerror());
    if (handle)
    {
        dlclose(handle);
        handle = 0;
    }
}

int upf_init_(const char *app_context)
{
    if (!handle)
    {
        handle = dlopen(UPF_CORE_NAME, RTLD_NOW | RTLD_LOCAL | RTLD_GROUP);
        if (!handle)
        {
            report_dlerror();
            return 0;
        }
        init = dlsym(handle, "upf_init_from_stub");
        if (!init)
        {
            report_dlerror();
            return 0;
        }
        done = dlsym(handle, "upf_done_impl");
        if (!done)
        {
            report_dlerror();
            return 0;
        }
    }
    if (!init(app_context, &upf_context, &upf_middleware))
    {
        dlclose(handle);
        handle = 0;
        return 0;        
    }

    return 1;
}

int upf_done(void)
{
    int rt = done();
    dlclose(handle);
    handle = 0;
    return rt;
}
