#ifndef __UPF_ENDIAN_H__
#define __UPF_ENDIAN_H__

// -------------------------------------------------------------------------
// Endianess detection:
// -------------------------------------------------------------------------

#ifdef HAVE_ENDIAN_H
  #include <endian.h>

  #if defined(__BYTE_ORDER) && defined(__LITTLE_ENDIAN) && defined(__BIG_ENDIAN)
    #if __BYTE_ORDER == __LITTLE_ENDIAN
      #define _UPF_LITTLE_ENDIAN
    #endif
    #if __BYTE_ORDER == __BIG_ENDIAN
      #define _UPF_BIG_ENDIAN
    #endif
  #endif
#endif

#if !defined(_UPF_LITTLE_ENDIAN) && !defined(_UPF_BIG_ENDIAN)
  // (Inspired by SDL and Boost checks)
  #if defined(__i386__) || defined(__ia64__) || defined(__x86_64__) || \
      defined(WIN32) || defined(__alpha__) || defined(__alpha) || \
      defined(__arm__) || (defined(__mips__) && defined(__MIPSEL__)) || \
      defined(__SYMBIAN32__) || defined(__LITTLE_ENDIAN__)
    #define _UPF_LITTLE_ENDIAN
  #elif defined(__sparc) || defined(__sparc__) || defined(__powerpc__) || \
        defined(__ppc__) || defined(__hppa) || defined(_MIPSEB) || \
        defined(__BIG_ENDIAN__)
    #define _UPF_BIG_ENDIAN
  #else
    #error "Please define endianess of your machine!"
  #endif
#endif

// -------------------------------------------------------------------------
// Swapping macros:
// -------------------------------------------------------------------------

#define _UPF_SWAP_16(x) \
      ((((x) & 0x00FFU) << 8) | (((x) & 0xFF00U) >> 8))

#define _UPF_SWAP_32(x) \
      ((((x) & 0x000000FFU) << 24) | \
       (((x) & 0x0000FF00U) <<  8) | \
       (((x) & 0x00FF0000U) >>  8) | \
       (((x) & 0xFF000000U) >> 24))


#ifdef _UPF_LITTLE_ENDIAN
    #define _UPF_SWAP_LE16(x) (x)
    #define _UPF_SWAP_LE32(x) (x)
    #define _UPF_SWAP_BE16(x) _UPF_SWAP_16(x)
    #define _UPF_SWAP_BE32(x) _UPF_SWAP_32(x)
#else
    #define _UPF_SWAP_LE16(x) _UPF_SWAP_16(x)
    #define _UPF_SWAP_LE32(x) _UPF_SWAP_32(x)
    #define _UPF_SWAP_BE16(x) (x)
    #define _UPF_SWAP_BE32(x) (x)
#endif

#define _UPF_SWAP_NET16(x) _UPF_SWAP_BE16(x)
#define _UPF_SWAP_NET32(x) _UPF_SWAP_BE32(x)


#endif
