
#include "upf/upf.h"

#include "boost/thread/thread.hpp"
#include "boost/thread/mutex.hpp"

#include "private.h"
#include "dll_loader.h"


using namespace upf;
using namespace std;

upf_middlewareLayer *upf_middleware = NULL;

namespace upf { namespace impl {
upf::impl::PrivateData *privateData;

} }

namespace
{

upf_middlewareLayer MW;
bool initialized = false;

void setupMiddlewareStruct()
{
    MW.ABI_version = UPF_ABI_VERSION;
    MW.manager = impl::privateData->manager->_upf_getABIObject();
    MW.serviceManager = NULL;

    MW.funcs.create_instance =        impl::create_instance;
    MW.funcs.marchaller_init =        impl::marchaller_init;
    MW.funcs.marchaller_seek0 =       impl::marchaller_seek0;
    MW.funcs.marchaller_destroy =     impl::marchaller_destroy;
    MW.funcs.marchaller_write =       impl::marchaller_write;
    MW.funcs.marchaller_read =        impl::marchaller_read;
    MW.funcs.get_context =            impl::get_context;
    MW.funcs.alloc_new_cid =          impl::alloc_new_cid;
    MW.funcs.classname_to_cid =       impl::classname_to_cid;
    MW.funcs.merge_typeinfo =         impl::merge_typeinfo;
}


#ifdef __UPF_UNIX__
const char *getModulesPath()
{
    const char *libdir = getenv("UPF_LIBDIR");
    if (libdir && strlen(libdir) > 0)
        return libdir;
    return UPF_LIBDIR;
}
#endif

}

#ifdef __UPF_UNIX__
    #define upf_init_ upf_init_impl
    #define upf_done  upf_done_impl
#endif

_UPF_EXPORTED_SYMBOL int upf_init_(const char *user_context)
{
    using impl::privateData;

    assert(!initialized);
    if (initialized) return 0;

    privateData = new impl::PrivateData;
    privateData->manager = new impl::Manager();
    if (privateData->manager == NULL)
        return 0;

    upf_middleware = &MW;
    setupMiddlewareStruct();

    impl::init_std_contexts(user_context);

    // finish manager creation - it has to be done manually because this is the
    // first class in existence:
    privateData->manager->_initClass();
    // must incRef here because previous command reseted ref count:
    privateData->manager->incRef();

    // Likewise, we must explicitly call ClassInfo's registrator before
    // registration will work with normal UPF API:
    impl::ClassInfo *ci_impl = new impl::ClassInfo();
    // Must initialize instance manually, factories not yet in place:
    ci_impl->_initInstance();
    Ptr<IWriteableClassInfo> ci = ci_impl;
    ClassRegistrator<impl::ClassInfo>::prepare(ci, "upf::impl::ClassInfo");
    Ptr<IClassInfo> frozen = ci->freeze();
    privateData->manager->registerClass(frozen);

    // Register the rest of internally implemented classes:
    UPF_REGISTER_CLASS(upf::impl::TypeInfo);
    UPF_REGISTER_CLASS(upf::impl::NativeTypeInfo);
    UPF_REGISTER_CLASS(upf::impl::TypedefInfo);
    UPF_REGISTER_CLASS(upf::impl::InterfaceInfo);
    UPF_REGISTER_CLASS(upf::impl::DllLoader);
    UPF_REGISTER_CLASS(upf::impl::ServiceManager);
    UPF_REGISTER_CLASS(upf::impl::DefaultLog);
    UPF_REGISTER_CLASS(upf::impl::LogChangeListener);
    
    // Initialize service manager and logging interfaces:
    privateData->serviceManager = create<IServiceManager>();
    MW.serviceManager = privateData->serviceManager->_upf_getABIObject();
    Ptr<IServiceChangeListener> listener = create("upf.impl.LogChangeListener");
    privateData->serviceManager->addListener(SERVICE_LOG, listener);
    Ptr<ILog> log = create("upf.impl.DefaultLog");
    privateData->serviceManager->bind(SERVICE_LOG,log, SERVICE_CAN_REBIND);

#if defined(__UPF_UNIX__)
    // Add default locations of runtime-loaded modules:
    // (NB: this may/will trigger modules loading)
    const char *basedir = getModulesPath();
    char version[12];
    snprintf(version, 12, "%i.%i", UPF_VERSION_MAJOR, UPF_VERSION_MINOR);
    privateData->manager->addModulesPath(string(basedir) + "/" +
                                         version  + "/loaders");
#endif

    initialized = true;
    return 1;
}

_UPF_EXPORTED_SYMBOL int upf_done(void)
{
    assert(initialized);
    if (!initialized) return 0;

    // NB: We must delete all private data first, and only then can the
    //     manager be deleted, because Manager's dtor unloads all DLLs and
    //     impl::privateData contains pointers to some UPF objects (e.g.
    //     services).
    impl::Manager *mng = impl::privateData->manager;
    delete impl::privateData;
    delete mng;
    
    impl::privateData = NULL;
    upf_middleware = NULL;
    initialized = false;
    return 1;
}

#ifdef __UPF_UNIX__
#undef upf_context
extern "C" int upf_init_from_stub(const char *app_context,
                                  upf_ContextID *context,
                                  upf_middlewareLayer **middleware)
{
    int rt = upf_init_(app_context);
    if (rt != 0)
    {
        *context = upf_context;
        *middleware = upf_middleware;
    }
    return rt;
}
#endif
