
#define UPF_BUILDING_MODULE

#include <upf/upf.h>
#include <upf/ILoader.h>

#include <iostream>
using namespace std;

#ifdef __GNUC__
   #define FakeLoader  FakeLoader_GCC
   #define FakeLoader2 FakeLoader2_GCC
#endif
#ifdef _MSC_VER
   #define FakeLoader  FakeLoader_MSVC
   #define FakeLoader2 FakeLoader2_MSVC
#endif
#ifdef __BORLANDC__
   #define FakeLoader  FakeLoader_Borland
   #define FakeLoader2 FakeLoader2_Borland
#endif
#ifdef __INTEL_COMPILER
   #define FakeLoader  FakeLoader_Intel
   #define FakeLoader2 FakeLoader2_Intel
#endif

#define CLNAME(name) _UPF_STRINGIZE(name)

class FakeLoader : public upf::ILoader
{
public:
    FakeLoader()
    {
        cout << CLNAME(FakeLoader) << ": constructed" << endl;
    }
    virtual ~FakeLoader()
    {
        cout << CLNAME(FakeLoader) << ": destroyed" << endl;
    }
 
    ExtensionsList getExtensions() { return ExtensionsList(); /*all files*/ }

    bool canLoad(const string& filename) { return true; }
    bool load(const string& filename);
    
    string getDescription()
        { return "fake loader that does not load anything"; }

    UPF_DECLARE_CLASS(FakeLoader)
};


bool FakeLoader::load(const string& filename)
{
    cout << CLNAME(FakeLoader) << ": loading classes from " << filename << endl;
    return true;
}


UPF_IMPLEMENT_CLASS(FakeLoader)
{
    UPF_INTERFACE(upf::ILoader)
    UPF_PROPERTY("PartOf", "UPF test suite")
    UPF_PROPERTY("Description", "This module only *pretends* it is a loader.")
}


class FakeLoader2 : public upf::ILoader
{
public:
    FakeLoader2()
    {
        cout << CLNAME(FakeLoader2) << ": constructed" << endl;
    }
    virtual ~FakeLoader2()
    {
        cout << CLNAME(FakeLoader2) << ": destroyed" << endl;
    }
 
    ExtensionsList getExtensions()
    {
        ExtensionsList list;
        list.push_back(".o");
        list.push_back(".so");
        list.push_back(".cpp");
        return list;
    }

    bool canLoad(const string& filename) { return true; }
    bool load(const string& filename);
    
    string getDescription()
        { return "fake loader II does nothing as well"; }

    UPF_DECLARE_CLASS(FakeLoader2)
};


bool FakeLoader2::load(const string& filename)
{
    cout << CLNAME(FakeLoader2) 
         << ": loading classes from " << filename << endl;
    return true;
}


UPF_IMPLEMENT_CLASS(FakeLoader2)
{
    UPF_INTERFACE(upf::ILoader)
    UPF_PROPERTY("PartOf", "UPF test suite")
    UPF_PROPERTY("Description", "This module only *pretends* it is a loader.")
}

static void onUnload()
{
    cout << "Unloading fakeloader..." << endl;
}

static void onLoad1()
{
    cout << "Loading fakeloader (void)..." << endl;
}

static bool onLoad2()
{
    cout << "Loading fakeloader (bool)..." << endl;
    return true;
}


UPF_DLL_MODULE()
{
    UPF_ON_UNLOAD(onUnload)
    UPF_ON_LOAD(onLoad1)
    UPF_ON_LOAD(onLoad2)

    UPF_EXPORTED_CLASS(FakeLoader)
    UPF_EXPORTED_CLASS(FakeLoader2)
}
