
#include <upf/upf.h>

#include <string>
#include <iostream>

#include "query.h"

using namespace std;
using namespace upf;

#include "utilities.h"

class MyImpl : public IfaceOne, public IfaceTwo
{
    MyImpl()
    {
       cout << "MyImpl constructed" << endl;
    }

    virtual ~MyImpl()
    {
       cout << "MyImpl destroyed" << endl;
    }

    void methodOne(const ::std::string& caller)
    {
       cout << "methodOne received: \"" << caller << "\"" << endl;
    }
    string methodTwo(IfaceOne *face)
    {
       return "methodTwo returned me";
    }

    UPF_DECLARE_CLASS(MyImpl)    
};

UPF_IMPLEMENT_CLASS(MyImpl)
{
    UPF_INTERFACE(IfaceOne)
    UPF_INTERFACE(IfaceTwo)
}


int main()
{
    upf::init();

    {
        UPF_REGISTER_CLASS(MyImpl);
    
        Ptr<IObject> subj = upf::create("MyImpl");
        
        if (subj == NULL)
            cerr << "error: cannot create MyImpl instance!" << endl;

        upf::CID cid = subj->getClassID();
        Ptr<IClassInfo> ci = upf::getManager()->getClassInfo(cid);
        cout << "MyImpl class: ";
        writeCID(cout, cid) << endl;

        cout << "  implements following interfaces:" << endl;
        IIDsList ifaces = ci->getInterfaces();
        for (IIDsList::iterator i = ifaces.begin(); i != ifaces.end(); i++)
        {
            cout << "    \"" << *i  << "\"" << endl;
        }
        
        Ptr<IfaceOne> one(subj);
        Ptr<IfaceTwo> two(subj);
        Ptr<upf::IManager> mng(subj);

        cout << "  casted to IfaceOne      : " << one << endl;
        cout << "  casted to IfaceTwo      : " << two << endl;
        cout << "  casted to upf::IManager : " << mng << endl;
        
        one->methodOne(two->methodTwo(one));        
    }

    // NB: all objects must be destroyed by now!
    upf::done();

    return 0;
}
