
#include <upf/upf.h>
#include <string>
#include <iostream>

using namespace std;
using namespace upf;

bool listIfaces = false;


void enumImpl()
{
    Ptr<IManager> mng = upf::getManager();

    const IID& iid = InterfaceTraits<IObject>::iid();
    CIDsList cis = mng->enumInterfaceImplementations(iid);

    for (CIDsList::iterator i = cis.begin(); i != cis.end(); i++)
    {
        Ptr<IClassInfo> ci(mng->getClassInfo(*i));
        cout << ci->getName() << endl;
        if (listIfaces)
        {
            IIDsList ifaces = ci->getInterfaces();
            for (IIDsList::iterator j = ifaces.begin(); j != ifaces.end(); j++)
            {
                cout << "    " <<
                        mng->getInterfaceInfo(*j)->getName() <<
                        endl;
            }
        }
    }
}


void usage(const char *argv0)
{
    fprintf(stderr,
            "Lists classes implemented by modules in directory\n"
            "Usage: %s [-i] dirs...\n\n"
            "       -i    list interfaces implemented by classes\n"
            "\n",
            argv0);
}


int main(int argc, char *argv[])
{
    upf::init();

    if (argc == 1)
    {
        usage(argv[0]);
        return 1;
    }

    for (int i = 1; i < argc; i++)
    {
        if (strcmp(argv[i], "-i") == 0)
            listIfaces = true;
    }
    
    {
        Ptr<IManager> mng(upf::getManager());
        for (int i = 1; i < argc; i++)
        {
            if (argv[i][0] != '-')
                mng->addModulesPath(argv[i]);
        }
    }

    enumImpl();
    
    upf::done();

    return 0;
}
